/////////////////////////////////////////////////////////////////////////
// $Id: mult8.cc,v 1.11 2002/10/25 11:44:35 bdenney Exp $
/////////////////////////////////////////////////////////////////////////
//
//  Copyright (C) 2001  MandrakeSoft S.A.
//
//    MandrakeSoft S.A.
//    43, rue d'Aboukir
//    75002 Paris - France
//    http://www.linux-mandrake.com/
//    http://www.mandrakesoft.com/
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA






#define NEED_CPU_REG_SHORTCUTS 1
#include "bochs.h"
#define LOG_THIS BX_CPU_THIS_PTR



  void
BX_CPU_C::MUL_ALEb(bxInstruction_c *i)
{
  Bit8u op2, op1;
  Bit16u product_16;
  bx_bool temp_flag;
  Bit8u op1_integrity, op2_integrity, result_integrity;
  Expression *Op1Expr, *Op2Expr, *LMulExpr, *HMulExpr;

  op1 = AL;
  op1_integrity = EAX_INTEGRITY;
#if BX_DEBUGGER
  if (BX_CPU_THIS_PTR sexecmode)
  {
  	Op1Expr = (Expression *) TheLabels.ReadSymbolicRegister(0, 0);
  }
#endif

  /* op2 is a register or memory reference */
  if (i->modC0()) {
    op2 = BX_READ_8BIT_REGx(i->rm(),i->extend8bitL());
    op2_integrity = BX_READ_8BIT_INTEGRITY(i->rm());
#if BX_DEBUGGER
  if (BX_CPU_THIS_PTR sexecmode)
  {
  	Op2Expr = (Expression *) TheLabels.ReadSymbolicRegister(i->rm(), 0);
  }
#endif
    }
  else {
    /* pointer, segment address pair */
    read_virtual_byte(i->seg(), RMAddr(i), &op2, &op2_integrity);
#if BX_DEBUGGER
  if (BX_CPU_THIS_PTR sexecmode)
  {
  	read_virtual_expression(i->seg(), RMAddr(i), &Op2Expr);
  }
#endif
    }

  product_16 = op1 * op2;
#if BX_DEBUGGER
  if (BX_CPU_THIS_PTR sexecmode)
  {
  	LMulExpr = (Expression *) new Operation(Op1Expr, Op2Expr, "LOW_MUL_ALEb");
  	HMulExpr = (Expression *) new Operation(Op1Expr, Op2Expr, "HIGH_MUL_ALEb");
  }
#endif

  result_integrity = low_water_mark(op1_integrity, op2_integrity, INTEG_BYTE, INTEG_BYTE);

  /* set EFLAGS:
   * MUL affects the following flags: C,O
   */

  temp_flag = ((product_16 & 0xFF00) != 0);
  SET_FLAGS_OxxxxC(temp_flag, temp_flag);

  /* now write product back to destination */

  AX = product_16;
  //LVAL_EAX_INTEGRITY = result_integrity;
  BX_WRITE_32BIT_REG_INTEGRITY(0, result_integrity);
#if BX_DEBUGGER
  if (BX_CPU_THIS_PTR sexecmode)
  {
	TheLabels.WriteSymbolicRegister(0, 0, LMulExpr);
	TheLabels.WriteSymbolicRegister(2, 0, HMulExpr);
	//if (Op1Expr != NULL) delete Op1Expr;
	//if (Op2Expr != NULL) delete Op2Expr;
  }
#endif
}


  void
BX_CPU_C::IMUL_ALEb(bxInstruction_c *i)
{
  Bit8s op2, op1;
  Bit16s product_16;
  Bit16u upper_bits;
  Bit8u op1_integrity, op2_integrity, result_integrity;
  Expression *Op1Expr, *Op2Expr, *LMulExpr, *HMulExpr;


  op1 = AL;
  op1_integrity = EAX_INTEGRITY;
#if BX_DEBUGGER
  if (BX_CPU_THIS_PTR sexecmode)
  {
  	Op1Expr = (Expression *) TheLabels.ReadSymbolicRegister(0, 0);
  }
#endif

  /* op2 is a register or memory reference */
  if (i->modC0()) {
    op2 = BX_READ_8BIT_REGx(i->rm(),i->extend8bitL());
    op2_integrity = BX_READ_8BIT_INTEGRITY(i->rm());
#if BX_DEBUGGER
  if (BX_CPU_THIS_PTR sexecmode)
  {
  	Op2Expr = (Expression *) TheLabels.ReadSymbolicRegister(i->rm(), 0);
  }
#endif
    }
  else {
    /* pointer, segment address pair */
    read_virtual_byte(i->seg(), RMAddr(i), (Bit8u *) &op2, &op2_integrity);
#if BX_DEBUGGER
  if (BX_CPU_THIS_PTR sexecmode)
  {
  	read_virtual_expression(i->seg(), RMAddr(i), &Op2Expr);
  }
#endif
    }

  product_16 = op1 * op2;
  result_integrity = low_water_mark(op1_integrity, op2_integrity, INTEG_BYTE, INTEG_BYTE);
#if BX_DEBUGGER
  if (BX_CPU_THIS_PTR sexecmode)
  {
  	LMulExpr = (Expression *) new Operation(Op1Expr, Op2Expr, "LOW_IMUL_ALEb");
  	HMulExpr = (Expression *) new Operation(Op1Expr, Op2Expr, "HIGH_IMUL_ALEb");
  }
#endif

  /* now write product back to destination */

  AX = product_16;

  //LVAL_EAX_INTEGRITY = result_integrity;
  BX_WRITE_32BIT_REG_INTEGRITY(0, result_integrity);
#if BX_DEBUGGER
  if (BX_CPU_THIS_PTR sexecmode)
  {
	TheLabels.WriteSymbolicRegister(0, 0, LMulExpr);
	TheLabels.WriteSymbolicRegister(2, 0, HMulExpr);
	//if (Op1Expr != NULL) delete Op1Expr;
	//if (Op2Expr != NULL) delete Op2Expr;
  }
#endif

  /* set EFLAGS:
   * IMUL affects the following flags: C,O
   * IMUL r/m8: condition for clearing CF & OF:
   *   AL = sign-extend of AL to 16 bits
   */
  upper_bits = AX & 0xff80;
  if (upper_bits==0xff80  ||  upper_bits==0x0000) {
    SET_FLAGS_OxxxxC(0, 0);
    }
  else {
    SET_FLAGS_OxxxxC(1, 1);
    }
}



  void
BX_CPU_C::DIV_ALEb(bxInstruction_c *i)
{
  Bit8u op2, quotient_8l, remainder_8;
  Bit16u quotient_16, op1;
  Bit8u op1_integrity, op2_integrity, result_integrity;
  Expression *Op1Expr, *Op2Expr, *LDivExpr, *RDivExpr;


  op1 = AX;
  op1_integrity = EAX_INTEGRITY;
#if BX_DEBUGGER
  if (BX_CPU_THIS_PTR sexecmode)
  {
  	Op1Expr = (Expression *) TheLabels.ReadSymbolicRegister(0, 0);
  }
#endif

  /* op2 is a register or memory reference */
  if (i->modC0()) {
    op2 = BX_READ_8BIT_REGx(i->rm(),i->extend8bitL());
    op2_integrity = BX_READ_8BIT_INTEGRITY(i->rm());
#if BX_DEBUGGER
  if (BX_CPU_THIS_PTR sexecmode)
  {
  	Op2Expr = (Expression *) TheLabels.ReadSymbolicRegister(i->rm(), 0);
  }
#endif
    }
  else {
    /* pointer, segment address pair */
    read_virtual_byte(i->seg(), RMAddr(i), &op2, &op2_integrity);
#if BX_DEBUGGER
  if (BX_CPU_THIS_PTR sexecmode)
  {
  	read_virtual_expression(i->seg(), RMAddr(i), &Op2Expr);
  }
#endif
    }

  if (op2 == 0) {
    exception(BX_DE_EXCEPTION, 0, 0);
    }
  quotient_16 = op1 / op2;
  remainder_8 = op1 % op2;
  quotient_8l = quotient_16 & 0xFF;
#if BX_DEBUGGER
  if (BX_CPU_THIS_PTR sexecmode)
  {
  	LDivExpr = (Expression *) new Operation(Op1Expr, Op2Expr, "QUOTIENT_DIV_ALEb");
  	RDivExpr = (Expression *) new Operation(Op1Expr, Op2Expr, "REMAINDER_DIV_ALEb");
  }
#endif

  if (quotient_16 != quotient_8l) {
    exception(BX_DE_EXCEPTION, 0, 0);
    }

  /* set EFLAGS:
   * DIV affects the following flags: O,S,Z,A,P,C are undefined
   */

#if INTEL_DIV_FLAG_BUG == 1
    set_CF(1);
#endif

  /* now write quotient back to destination */

  AL = quotient_8l;
  AH = remainder_8;
  result_integrity = low_water_mark(op1_integrity, op2_integrity, INTEG_WORD, INTEG_BYTE);
  //LVAL_EAX_INTEGRITY = result_integrity;
  BX_WRITE_32BIT_REG_INTEGRITY(0, result_integrity);
#if BX_DEBUGGER
  if (BX_CPU_THIS_PTR sexecmode)
  {
	TheLabels.WriteSymbolicRegister(0, 0, LDivExpr);
	TheLabels.WriteSymbolicRegister(2, 0, RDivExpr);
	//if (Op1Expr != NULL) delete Op1Expr;
	//if (Op2Expr != NULL) delete Op2Expr;
  }
#endif
}


  void
BX_CPU_C::IDIV_ALEb(bxInstruction_c *i)
{
  Bit8s op2, quotient_8l, remainder_8;
  Bit16s quotient_16, op1;
  Bit8u op1_integrity, op2_integrity, result_integrity;
  Expression *Op1Expr, *Op2Expr, *LDivExpr, *RDivExpr;


  op1 = AX;
  op1_integrity = EAX_INTEGRITY;
#if BX_DEBUGGER
  if (BX_CPU_THIS_PTR sexecmode)
  {
  	Op1Expr = (Expression *) TheLabels.ReadSymbolicRegister(0, 0);
  }
#endif


  /* op2 is a register or memory reference */
  if (i->modC0()) {
    op2 = BX_READ_8BIT_REGx(i->rm(),i->extend8bitL());
    op2_integrity = BX_READ_8BIT_INTEGRITY(i->rm());
#if BX_DEBUGGER
  if (BX_CPU_THIS_PTR sexecmode)
  {
  	Op2Expr = (Expression *) TheLabels.ReadSymbolicRegister(i->rm(), 0);
  }
#endif
    }
  else {
    /* pointer, segment address pair */
    read_virtual_byte(i->seg(), RMAddr(i), (Bit8u *) &op2, &op2_integrity);
#if BX_DEBUGGER
  if (BX_CPU_THIS_PTR sexecmode)
  {
  	read_virtual_expression(i->seg(), RMAddr(i), &Op2Expr);
  }
#endif
    }

  if (op2 == 0) {
    exception(BX_DE_EXCEPTION, 0, 0);
    }

  quotient_16 = op1 / op2;
  remainder_8 = op1 % op2;
  quotient_8l = quotient_16 & 0xFF;
#if BX_DEBUGGER
  if (BX_CPU_THIS_PTR sexecmode)
  {
  	LDivExpr = (Expression *) new Operation(Op1Expr, Op2Expr, "QUOTIENT_IDIV_ALEb");
  	RDivExpr = (Expression *) new Operation(Op1Expr, Op2Expr, "REMAINDER_IDIV_ALEb");
  }
#endif

  if (quotient_16 != quotient_8l) {
BX_INFO(("quotient_16: %04x, remainder_8: %02x, quotient_8l: %02x",
  (unsigned) quotient_16, (unsigned) remainder_8, (unsigned) quotient_8l));
AL = quotient_8l;
AH = remainder_8;
BX_INFO(("AH: %02x, AL: %02x", (unsigned) AH, (unsigned) AL));
    exception(BX_DE_EXCEPTION, 0, 0);
    }

  /* set EFLAGS:
   * DIV affects the following flags: O,S,Z,A,P,C are undefined
   */

#if INTEL_DIV_FLAG_BUG == 1
    set_CF(1);
#endif

  /* now write quotient back to destination */

  AL = quotient_8l;
  AH = remainder_8;
  result_integrity = low_water_mark(op1_integrity, op2_integrity, INTEG_WORD, INTEG_BYTE);
  //LVAL_EAX_INTEGRITY = result_integrity;
  BX_WRITE_32BIT_REG_INTEGRITY(0, result_integrity);
#if BX_DEBUGGER
  if (BX_CPU_THIS_PTR sexecmode)
  {
	TheLabels.WriteSymbolicRegister(0, 0, LDivExpr);
	TheLabels.WriteSymbolicRegister(2, 0, RDivExpr);
	//if (Op1Expr != NULL) delete Op1Expr;
	//if (Op2Expr != NULL) delete Op2Expr;
  }
#endif
}
